/*
  CatWeasel MK3 PCI Linux driver (c) 2002 by Dirk Jagdmann <doj@cubic.org>

  This file is part of CatWeasel MK3 PCI Linux driver.

  The CatWeasel MK3 PCI Linux driver is free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  The CatWeasel MK3 PCI Linux driver is distributed in the hope that
  it will be useful, but WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the CatWeasel MK3 PCI Linux driver; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
  MA 02111-1307 USA

  $Header: /home/cvs/sidblaster/device/cwjoy.c,v 1.3 2003/06/01 13:43:12 doj Exp $
*/

/*
  Code for accessing the joysticks and paddles
*/

#include <cw.h>

#include <asm/io.h>
#include <linux/delay.h>
#include <linux/input.h>
#include <linux/kbd_ll.h>
#include <linux/timer.h>

static int cwjoy_rate=25;
MODULE_PARM(cwjoy_rate,"i");
MODULE_PARM_DESC(cwjoy_rate,"update joysticks and paddles x times per second (default=25)");

static int keytab[128];

static int cwkey_delay=HZ/3;
MODULE_PARM(cwkey_delay, "i");
MODULE_PARM_DESC(cwkey_delay, "delay amiga keyboard repeat by jiffies");

static int cwkey_translation=1;
MODULE_PARM(cwkey_translation, "i");
MODULE_PARM_DESC(cwkey_translation, "enable/disable keyboard translation (default 1=enabled)");

#define LINUX_INPUT_HAS_BTN_BUG 1

enum {
  Joystick1 = 0,
  Joystick2,
  Paddle1,
  Paddle2,
  Nummer,
};

static struct input_dev cwjoy_dev[CWSID_MAX_NUM][Nummer];

static void cwjoy_poll(unsigned long l);

static struct timer_list cwjoy_timer = {
  function: cwjoy_poll,
  data: 0
};

static void sendscancode(int sc)
{
  if((sc&0xff00) == 0xe000)
    handle_scancode(0xe0, !(sc&0x80));
  handle_scancode(sc&0x7f, !(sc&0x80));
}

static void cwjoy_poll(unsigned long l)
{
  int i;

  /* reschedule ourself */
  cwjoy_timer.expires=jiffies+HZ/cwjoy_rate;
  add_timer(&cwjoy_timer);

  for(i=0; i<cw_num; i++)
    {
      /* joysticks */
      const unsigned char d=inb(cw_register[i][JoyDat]);
      const unsigned char b=inb(cw_register[i][JoyButton]);

      input_report_abs(&cwjoy_dev[i][Joystick1], ABS_X, (d&1?0:1) - (d&2?0:1));
      input_report_abs(&cwjoy_dev[i][Joystick1], ABS_Y, (d&4?0:1) - (d&8?0:1));
      input_report_key(&cwjoy_dev[i][Joystick1], BTN_TRIGGER, !(b&64));

      input_report_abs(&cwjoy_dev[i][Joystick2], ABS_X, (d&16?0:1) - (d&32?0:1));
      input_report_abs(&cwjoy_dev[i][Joystick2], ABS_Y, (d&64?0:1) - (d&128?0:1));
      input_report_key(&cwjoy_dev[i][Joystick2], BTN_TRIGGER, !(b&128));

      /* paddles */
      static int toggle=0;
      if((toggle=!toggle))
	{
	  input_report_abs(&cwjoy_dev[i][Paddle1], ABS_X, sid_read(0x19, i));
	  input_report_abs(&cwjoy_dev[i][Paddle1], ABS_Y, sid_read(0x1A, i));
	  outb(0, cw_register[i][PaddleSelect]); /* switch to paddle 2 */
	}
      else
	{
	  input_report_abs(&cwjoy_dev[i][Paddle2], ABS_X, sid_read(0x19, i));
	  input_report_abs(&cwjoy_dev[i][Paddle2], ABS_Y, sid_read(0x1A, i));
	  inb(cw_register[i][PaddleSelect]); /* switch to paddle 1 */
	}

      /* amiga keyboard */
      static int needunlock=0;
      static int sc=0;
      static unsigned long keydelay=~0;
      if(needunlock==1)
	{
	  outb(0, cw_register[i][KeyDat]);
	  needunlock=0;
	}
      else if(inb(cw_register[i][KeyStatus]) & 0x80)
	{
	  needunlock=1;		/* we delay unlocking the catweasel by one timer call */

	  const unsigned char key=inb(cw_register[i][KeyDat]); /* read key */
	  int newsc=keytab[key&0x7f];

	  /* is this scancode valid? */
	  if(newsc>0)
	    {
	      newsc|=(key&0x80);	/* save up/down information in scancode */
	      sendscancode(newsc);

	      /* a keyup event ? */
	      if(newsc&0x80)
		{
		  /* if this keyup is the repeating keydown, we remove keyrepeat */
		  if(newsc==(sc|0x80))
		    {
		      sc=0;
		      keydelay=~0;
		    }
		}
	      /* keydown event */
	      else
		{
		  sc=newsc;
		  keydelay=jiffies+cwkey_delay;
		}
	    }
#if 0  
	  if(!(key&0x80))
	    printk(KERN_INFO DEVICE_NAME ": key %i\n", key);
#endif
	}
      /* make software keyrepeat */
      if(sc>0 && jiffies>keydelay)
	sendscancode(sc);

    } /* for(i */
}

int __init cwjoy_init()
{
  int i;

  static char str[CWSID_MAX_NUM][4][32];

  for(i=0; i<cw_num; i++)
    {
      sprintf(str[i][0], "CatWeasel %i Joystick 1", i);
      cwjoy_dev[i][Joystick1].name = str[i][0];
      cwjoy_dev[i][Joystick1].idbus = BUS_PCI;
      cwjoy_dev[i][Joystick1].idvendor = 0xE159;
      cwjoy_dev[i][Joystick1].idproduct = 1;
      cwjoy_dev[i][Joystick1].idversion = 0x0100;
      cwjoy_dev[i][Joystick1].evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
      cwjoy_dev[i][Joystick1].absbit[0] = BIT(ABS_X) | BIT(ABS_Y);
      set_bit(BTN_TRIGGER, cwjoy_dev[i][Joystick1].keybit);
      cwjoy_dev[i][Joystick1].absmin[ABS_X] = -1; cwjoy_dev[i][Joystick1].absmax[ABS_X] = 1;
      cwjoy_dev[i][Joystick1].absmin[ABS_Y] = -1; cwjoy_dev[i][Joystick1].absmax[ABS_Y] = 1;
      input_register_device(&cwjoy_dev[i][Joystick1]);

      sprintf(str[i][1], "CatWeasel %i Joystick 2", i);
      cwjoy_dev[i][Joystick2].name = str[i][1];
      cwjoy_dev[i][Joystick2].idbus = BUS_PCI;
      cwjoy_dev[i][Joystick2].idvendor = 0xE159;
      cwjoy_dev[i][Joystick2].idproduct = 1;
      cwjoy_dev[i][Joystick2].idversion = 0x0100;
      cwjoy_dev[i][Joystick2].evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
      cwjoy_dev[i][Joystick2].absbit[0] = BIT(ABS_X) | BIT(ABS_Y);
      set_bit(BTN_TRIGGER, cwjoy_dev[i][Joystick2].keybit);
      cwjoy_dev[i][Joystick2].absmin[ABS_X] = -1; cwjoy_dev[i][Joystick2].absmax[ABS_X] = 1;
      cwjoy_dev[i][Joystick2].absmin[ABS_Y] = -1; cwjoy_dev[i][Joystick2].absmax[ABS_Y] = 1;
      input_register_device(&cwjoy_dev[i][Joystick2]);

      sprintf(str[i][2], "CatWeasel %i Paddle 1", i);
      cwjoy_dev[i][Paddle1].name = str[i][2];
      cwjoy_dev[i][Paddle1].idbus = BUS_PCI;
      cwjoy_dev[i][Paddle1].idvendor = 0xE159;
      cwjoy_dev[i][Paddle1].idproduct = 1;
      cwjoy_dev[i][Paddle1].idversion = 0x0100;
#if LINUX_INPUT_HAS_BTN_BUG
      cwjoy_dev[i][Paddle1].evbit[0] = BIT(EV_ABS) | BIT(EV_KEY);
      set_bit(BTN_TRIGGER, cwjoy_dev[i][Paddle1].keybit);
#else
      cwjoy_dev[i][Paddle1].evbit[0] = BIT(EV_ABS);
#endif
      cwjoy_dev[i][Paddle1].absbit[0] = BIT(ABS_X) | BIT(ABS_Y);
      cwjoy_dev[i][Paddle1].absmin[ABS_X] = 0; cwjoy_dev[i][Paddle1].absmax[ABS_X] = 0xFF;
      cwjoy_dev[i][Paddle1].absmin[ABS_Y] = 0; cwjoy_dev[i][Paddle1].absmax[ABS_Y] = 0xFF;
      input_register_device(&cwjoy_dev[i][Paddle1]);

      sprintf(str[i][3], "CatWeasel %i Paddle 2", i);
      cwjoy_dev[i][Paddle2].name = str[i][3];
      cwjoy_dev[i][Paddle2].idbus = BUS_PCI;
      cwjoy_dev[i][Paddle2].idvendor = 0xE159;
      cwjoy_dev[i][Paddle2].idproduct = 1;
      cwjoy_dev[i][Paddle2].idversion = 0x0100;
#if LINUX_INPUT_HAS_BTN_BUG
      cwjoy_dev[i][Paddle2].evbit[0] = BIT(EV_ABS) | BIT(EV_KEY);
      set_bit(BTN_TRIGGER, cwjoy_dev[i][Paddle2].keybit);
#else
      cwjoy_dev[i][Paddle2].evbit[0] = BIT(EV_ABS);
#endif
      cwjoy_dev[i][Paddle2].absbit[0] = BIT(ABS_X) | BIT(ABS_Y);
      cwjoy_dev[i][Paddle2].absmin[ABS_X] = 0; cwjoy_dev[i][Paddle2].absmax[ABS_X] = 0xFF;
      cwjoy_dev[i][Paddle2].absmin[ABS_Y] = 0; cwjoy_dev[i][Paddle2].absmax[ABS_Y] = 0xFF;
      input_register_device(&cwjoy_dev[i][Paddle2]);

      /* enable amiga keyboard */
      outb(0, cw_register[i][KeyStatus]);
      outb(0, cw_register[i][KeyDat]);
    }

  /* enable timer */
  init_timer(&cwjoy_timer);
  cwjoy_timer.expires=jiffies+HZ/cwjoy_rate;
  add_timer(&cwjoy_timer);

  /* set up scancode conversion table */
  if(cwkey_translation)
    {
      for(i=0; i<128; i++)
	keytab[i]=-1;

      keytab[69]=0x01;		/* Esc */
      keytab[80]=0x3b;		/* f1-f10 */
      keytab[81]=0x3c;
      keytab[82]=0x3d;
      keytab[83]=0x3e;
      keytab[84]=0x3f;
      keytab[85]=0x40;
      keytab[86]=0x41;
      keytab[87]=0x42;
      keytab[88]=0x43;
      keytab[89]=0x44;

      keytab[0]=0x29;		/* '~ */
      keytab[1]=0x02;		/* 1-0 */
      keytab[2]=0x03;
      keytab[3]=0x04;
      keytab[4]=0x05;
      keytab[5]=0x06;
      keytab[6]=0x07;
      keytab[7]=0x08;
      keytab[8]=0x09;
      keytab[9]=0x0a;
      keytab[10]=0x0b;
      keytab[11]=0x0c;		/* ?\ */
      keytab[12]=0x0d;		/* '`+= */
      keytab[13]=0x0e;		/* |\ */
      keytab[65]=0x0e;		/* backspace */

      keytab[66]=0x0f;		/* tab */
      keytab[16]=0x10;		/* q- */
      keytab[17]=0x11;
      keytab[18]=0x12;
      keytab[19]=0x13;
      keytab[20]=0x14;
      keytab[21]=0x15;
      keytab[22]=0x16;
      keytab[23]=0x17;
      keytab[24]=0x18;
      keytab[25]=0x19;
      keytab[26]=0x1a;
      keytab[27]=0x1b;		/* +*~ */
      keytab[68]=0x1c;		/* enter */

      keytab[99]=0x1d;		/* ctrl */
      keytab[98]=0x3a;		/* caps lock */
      keytab[32]=0x1e;		/* a- */
      keytab[33]=0x1f;
      keytab[34]=0x20;
      keytab[35]=0x21;
      keytab[36]=0x22;
      keytab[37]=0x23;
      keytab[38]=0x24;
      keytab[39]=0x25;
      keytab[40]=0x26;
      keytab[41]=0x27;
      keytab[42]=0x28;
      keytab[43]=0x29;		/* #' */

      keytab[96]=0x2a;		/* leftshift */
      keytab[48]=0x56;		/* <>| */
      keytab[49]=0x2c;		/* y-m */
      keytab[50]=0x2d;
      keytab[51]=0x2e;
      keytab[52]=0x2f;
      keytab[53]=0x30;
      keytab[54]=0x31;
      keytab[55]=0x32;
      keytab[56]=0x33;		/* . */
      keytab[57]=0x34;		/* , */
      keytab[58]=0x35;		/* - */
      keytab[97]=0x36;		/* rightshift */

      keytab[100]=0x38;		/* leftalt */
      /*keytab[102]=;*/		/* leftamiga */
      keytab[64]=0x39;		/* space */
      /*keytab[103]=;*/		/* rightamiga */
      keytab[101]=0xe038;	/* rightalt */

      keytab[70]=0xe053;	/* delete */
      keytab[95]=0xe052;	/* help */

      keytab[76]=0xe048;	/* up */
      keytab[79]=0xe04b;	/* left */
      keytab[77]=0xe050;	/* down */
      keytab[78]=0xe04d;	/* right */

      keytab[90]=0x45;		/* numlock */
      keytab[91]=0x46;		/* scrolllock */
      keytab[92]=0xe035;	/* kpslash */
      keytab[93]=0x37;		/* kpasterix */

      keytab[61]=0x47;		/* kp7 */
      keytab[62]=0x48;		/* kp8 */
      keytab[63]=0x49;		/* kp9 */
      keytab[74]=0x4e;		/* kp- */

      keytab[45]=0x4b;		/* kp4 */
      keytab[46]=0x4c;		/* kp5 */
      keytab[47]=0x4d;		/* kp6 */
      keytab[94]=0x4e;		/* kp+ */

      keytab[29]=0x4f;		/* kp1 */
      keytab[30]=0x50;		/* kp2 */
      keytab[31]=0x51;		/* kp3 */

      keytab[15]=0x52;		/* kp0 */
      keytab[60]=0x53;		/* kp. */
      keytab[67]=0xe01c;	/* kpenter */
    }
  else
    for(i=0; i<128; i++)
      keytab[i]=i;

  return 0;
}

void __exit cwjoy_exit()
{
  int c;
  for(c=0; c<cw_num; c++)
    {
      int i;
      for(i=0; i<Nummer; i++)
	input_unregister_device(&cwjoy_dev[c][i]);
    }
  del_timer_sync(&cwjoy_timer);
}
