/*
  sid (c) 1999-2003 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /home/cvs/sidblaster/sid/6581.cc,v 1.8 2003/05/22 23:21:41 doj Exp $
*/

#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "6581.h"
#include <cwsid.h>

MOS6581::MOS6581(const int n) :
  num(n),
  fh(-1),
  Filter(true),
  irqAvailable_(false),
  pal_(true)
{
  memset(buf, 0, sizeof(buf));
  Voice[1]=Voice[2]=Voice[3]=true;
}

MOS6581::~MOS6581()
{
  if(fh>=0)
    {
      reset();
      ::close(fh);
    }
}

const int MOS6581::open(const char *dev)
{
  if(fh>=0)
    {
      reset();
      ::close(fh);
      fh=-1;
    }

  f_pos=-1;
  if(dev)
    fh=::open(dev, O_RDWR);
  else
    {
      fh=::open("/dev/sid", O_RDWR);
      if(fh<0)
	fh=::open("/dev/misc/sid", O_RDWR);
    }

  // check if this is a CatWeasel style SID device
  if(fh>=0)
    {
      int i;
      if(ioctl(fh, CWSID_IOCTL_CARDTYPE, &i) < 0)
	{
	  ::close(fh);
	  fh=-1;
	}
      else if(i!=CWSID_MAGIC)
	{
	  ::close(fh);
	  fh=-1;
	}
    }

  // check if IRQs are available
  if(fh>=0)
    {
      int i=0;
      ioctl(fh, CWSID_IOCTL_IRQAVAIL, &i);
      if(i)
	irqAvailable_=true;
    }

  return fh;
}


void MOS6581::mute()
{
  for(int i=1; i<=3; i++)
    if(Voice[i])
      mute(i);
}

void MOS6581::unmute()
{
  for(int i=1; i<=3; i++)
    if(!Voice[i])
      unmute(i);
  if(!Filter)
    unmute(0);
}

/// see http://sourceforge.net/docman/display_doc.php?docid=1212&group_id=9266
void MOS6581::mute(int i)
{
  switch(i)
    {
    case 1:
      gateMute[1]=buf[4]&1;
      buf[4]&=0xFE;
      Voice[1]=false;
      i=4;
      break;
    case 2:
      gateMute[2]=buf[0xB]&1;
      buf[0xB]&=0xFE;
      Voice[2]=false;
      i=0xB;
      break;
    case 3:
      gateMute[3]=buf[0x18]&0x80;
      buf[0x18]|=0x80;
      write(0x18);
      gateMute[0]=buf[0x17]&4;
      buf[0x17]&=0xFB;
      i=0x17;
      Voice[3]=false;
      break;
    case 0:
      filter=buf[0x17]&0xF;
      buf[0x17]&=0xF0;
      Filter=false;
      i=0x17;
      break;
    default:
      return;
    }

  write(i);
}

void MOS6581::unmute(int i)
{
  switch(i)
    {
    case 1:
      if(gateMute[1])
	buf[4]|=1;
      Voice[1]=true;
      i=4;
      break;
    case 2:
      if(gateMute[2])
	buf[0xB]|=1;
      Voice[2]=true;
      i=0xB;
      break;
    case 3:
      if(gateMute[3])
	buf[0x18]|=128;
      else
	buf[0x18]&=0x7F;
      write(0x18);
      if(gateMute[0])
	buf[0x17]|=4;
      i=0x17;
      Voice[3]=true;
      break;
    case 0:
      buf[0x17]|=filter;
      Filter=true;
      i=0x17;
      break;
    default:
      return;
    }

  write(i);
}

const bool MOS6581::toggle(const int v)
{
  if(v<1 || v>3)
    return false;

  if((Voice[v]=!Voice[v]))
    unmute(v);
  else
    mute(v);
  return Voice[v];
}

const bool MOS6581::toggleFilter()
{ 
  if((Filter=!Filter))
    unmute(0);
  else
    mute(0);
  return Filter;
}

void MOS6581::set(const int i, byte b) 
{ 
  if(i<0 || i>=MAX)
    return;

  switch(i)
    {
    case 0: setfreq(1, b, realfreqhi[1]); return;
    case 1: setfreq(1, realfreqlo[1], b); return;

    case 4:
      // check if voice1 is muted
      if(!Voice[1])
	b&=0xFE;
      // check if gate is triggered
      if(!(buf[4]&1) && (b&1))
	gateTriggered_[1]=true;
      break;

    case 7: setfreq(2, b, realfreqhi[2]); return;
    case 8: setfreq(2, realfreqlo[2], b); return;

    case 0xB:
      // check if voice2 is muted
      if(!Voice[2])
	b&=0xFE;
      // check if gate is triggered
      if(!(buf[0xB]&1) && (b&1))
	gateTriggered_[2]=true;
      break;

    case 0xE: setfreq(3, b, realfreqhi[3]); return;
    case 0xF: setfreq(3, realfreqlo[3], b); return;

    case 0x12:
      // check if gate is triggered
      if(Voice[3] && !(buf[0x12]&1) && (b&1))
	gateTriggered_[3]=true;
      break;

    case 0x17:
      // check if voice3 is muted
      if(!Filter)
	b&=0xF0;
      else if(!Voice[3])
	b&=0xFB;
      break;

    case 0x18:
      // check if voice2 is muted
      if(!Voice[3])
	b|=0x80;
      // check if louder than maxvolume
      realvolume=b&0xF;
      if(realvolume>maxvolume)
	{
	  b&=0xF0;
	  b|=maxvolume&0xF;
	}
      break;
    }

  buf[i]=b;
  write(i);
}

void MOS6581::write(const int i)
{
  if(fh >= 0)
    {
      seek(i);
      ::write(fh, buf+i, 1);
      f_pos++;
    }
}

void MOS6581::setfreq(const int v, const byte lo, const byte hi)
{
  realfreqlo[v]=lo;
  realfreqhi[v]=hi;

  word freq=lohi(lo, hi);
  int f=static_cast<int>(static_cast<double>(freq)*transpose_);
  if(f<=0xFFFF)
    freq=f;
  else
    freq=0xFFFF;

  switch(v)
    {
    case 1: buf[0]=freq&0xFF; buf[1]=freq>>8; write(0); write(1); break;
    case 2: buf[7]=freq&0xFF; buf[8]=freq>>8; write(7); write(8); break;
    case 3: buf[0xE]=freq&0xFF; buf[0xF]=freq>>8; write(0xE); write(0xF); break;
    }
}

byte MOS6581::get(const int i)
{
  if(i<0 || i>=MAX)
    return 0;

  if(i>=0x19 && i<=0x1C)
    {
      seek(i);
      ::read(fh, buf+i, 1);
      f_pos++;
    }
  return buf[i];
}

void MOS6581::reset()
{
  transpose_=1;
  maxvolume=15;
  realfreqlo[1]=realfreqhi[1];
  realfreqlo[2]=realfreqhi[2];
  realfreqlo[3]=realfreqhi[3];
  for(int i=0; i<0x19; i++)
    set(i, 0);
}

const bool MOS6581::ntsc()
{
  pal_=false;
  if(fh>=0)
    return ioctl(fh, CWSID_IOCTL_NTSC)==0;
  return false;
}

const bool MOS6581::pal()
{
  pal_=true;
  if(fh>=0)
    return ioctl(fh, CWSID_IOCTL_PAL)==0;
  return false;
}

void MOS6581::set(SID& s)
{
  memcpy(buf, &s, sizeof(SID));
  for(int i=0; i<=0x18; i++)
    set(i, buf[i]);
}

void MOS6581::get(SID& s)
{ 
  if(fh>=0)
    {
      seek(0x19);
      ::read(fh, buf+0x19, 4);
      f_pos+=4;
    }
  memcpy(&s, buf, sizeof(SID));
}

void MOS6581::seek(int pos)
{
  if(fh>=0)
    {
      if(f_pos!=pos)
	{
	  lseek(fh, pos+num*0x20, SEEK_SET);
	  f_pos=pos;
	}
    }
}

void MOS6581::setVolume(const int v)
{
  set(0x18, (buf[0x18]&0xF0)|(v&0x0F));
}

void MOS6581::maxVolume(const int v)
{
  if(v>=0 && v<=15)
    {
      // set realvolume again, so it can be adjusted with the set() further down
      buf[0x18]&=0xF0;
      buf[0x18]|=realvolume;

      maxvolume=v;
      set(0x18, buf[0x18]);
    }
}

bool MOS6581::voiceActive(const int v)
{
  switch(v)
    {
    case 0: return Filter;
    case 1: return Voice[1];
    case 2: return Voice[2];
    case 3: return Voice[3];
    }
  return false;
}

void MOS6581::setGateBits(const bool b)
{
  gateTriggered_[1]=gateTriggered_[2]=gateTriggered_[3]=b;
}

bool MOS6581::gateTriggered(const int v)
{
  switch(v)
    {
    case 1: return gateTriggered_[1];
    case 2: return gateTriggered_[2];
    case 3: return gateTriggered_[3];
    }
  return false;
}

int MOS6581::irqDelay(const int delay)
{
  if(fh<0)
    return -1;
  int hz=(pal_?985248:1022700)/delay;
  if(ioctl(fh, CWSID_IOCTL_SETHZ, hz) < 0)
    return -1;
  return 0;
}
