/*
  sid (c) 1999-2003 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /home/cvs/sidblaster/sid/c64.h,v 1.13 2003/05/22 23:21:41 doj Exp $
*/

#ifndef C64__H
#define C64__H

#include <set>

#include "defines.h"
#include "sid.h"

class MOS6510;
class MOS6581;

/// a C64 Computer
class C64
{
 private:
  C64(const C64&);
  C64& operator= (const C64&);

 protected:
  /// memory
  byte *mem;
  
  /// cpu
  MOS6510 *cpu_;

  /// sids
  MOS6581* sid[CWSID_MAX_NUM];
  /// base addresses of sids
  int sidBase[CWSID_MAX_NUM];
  /// number of currently used sids. should be 0<sidNum_<CWSID_MAX_NUM
  int sidNum_;

  /// true if C64 is clocked with PAL frequency. false if NTSC
  bool pal_;

  /// true if pause between calls to playAddr should be calculated via CIA1 Timer1. False if 50/60Hz should be used.
  bool CIAdelay;

  /// container type for memory watchpoints
  typedef std::set<word> watchpoint_t;
  /// list of memory watchpoints
  watchpoint_t watchpoint;

 public:

  /// Frequency of C64 in PAL mode
  static const int palfreq;
  /// Frequency of C64 in NTSC mode
  static const int ntscfreq;

  /// make a C64
  C64();

  virtual ~C64();

  /** get a byte from memory. May toggle a watchpoint.
      @param addr address to use
      @return the content of addr
  */
  const byte getmem(const word addr);

  /** set a byte in memory. May toggle a watchpoint.
      @param addr address to use
      @param value byte to set
  */
  void setmem(const word addr, const byte value);

  /// @return pointer to begin of memory
  byte* memory() const { return mem; }

  /** lets you execute a subroutine starting in memory at npc.
      jsr performs a cpu reset and pushes two 0 onto the stack. It then jumps
      into the routine starting at npc. Because the stack contains two zeros
      the subroutine can exit either via a RTS instruction.

      @param npc the content of the program counter (pc)
      @param na the content of the accumulator (a)
      @param nx the content of the x-register (x)
      @param ny the content of the y-register (y)
      @param maxCycles maximum number of cycles the CPU should run. -1 for unlimited.
      @return the number of cycles used by the executed routine
  */
  const int jsr(const word npc=0, const byte na=0, const byte nx=0, const byte ny=0, const int maxCycles=1) const;

  /// performs a reset
  void reset();

  /// resets all SIDs
  void resetSID();

  /**
     set frequency.
     @param b true for PAL, false for NTSC
  */
  void pal(const bool b);

  /// mute all SIDs
  void mute();

  /// unmute all SIDs
  void unmute();

  /**
     toggle SID voice v.
     @param v SID voice to toggle. 0<v<sidNum_*3
     @return new state of toggled voice
  */
  bool toggle(const int v);

  /**
     @param v SID voice to query. 0<v<sidNum_*3
     @return true if voice is active, thus not muted.
  */
  bool voiceActive(const int v);

  /**
     set unix device SID chips should use.
     @param device a filename to a unix device. 0 if standard devices should be used.
     @return 0 on success. -1 on error.
  */
  int setSIDdevice(const char* device=0);

  /// set all SID registers to a
  void set(allSID& a);

  /// copy all SID registers to a
  void get(allSID& a);

  /// set volume on all SIDs to v.
  void setVolume(const int v=15);

  /// set maximum volume on all SIDs to v.
  void maxVolume(const int v=15);

  /**
     set base address to SID s to b
     @param s SID number. 0<=s<CWSID_MAX_NUM.
     @param b base address. 0<=b<=$FFFF, -1 disables the SID.
  */
  void setBase(const int s, const int b);

  /// @return number of currently active SIDs
  int sidNum() const { return sidNum_; }

  /// @return true if C64 is PAL clocked, false for NTSC
  bool getPal() const { return pal_; }

  /// set gate detection bits in all SIDs to b
  void setGateBits(const bool b=false);

  /**
     query gate detection bits. The gate detection bits are cleared by setGateBits()
     @param v SID voice which should be queued. 0<v<sidNum_*3
     @return true if ADSR in voice has been started, false if not.
  */
  bool gateTriggered(const int v);

  /// @return current delay value in us.
  int delay();

  /// set registers in CIA1 Timer1 according to current frequency of C64
  void setCIAtimer1();

  /// enable or disable usage of CIA1 Timer1 for delay calculation depending of b.
  void setCIAdelay(const bool b);

  /// @return cpu of C64
  MOS6510* cpu() { return cpu_; }

  /**
     enable or disable a memory watchpoint at addr.
  */
  void watchToggle(const word addr);
  /// clear all memory watchpoints
  void watchClear();
  /// print all memory watchpoints to stdout
  void watchPrint();
 protected:
  /// @return true if a watchpoint exists for addr.
  bool watch(const word addr);
 public:

  /// @return unix file handle for unix device used by SIDs
  int sidfh() const;

  /// @return true if SIDs provide IRQ enabled delaying/blocking
  bool irqAvailable() const;

  /**
     set delay of IRQ to delay us.
     @param delay delay value in us.
     @return 0 on success, -1 on error.
  */
  int irqDelay(const int delay);

  /// set transpose of all SIDs to t
  void transpose(const double t);
};

#endif
